# -*- coding: utf-8 -*-
"""
Resources below '/<api_base>/websocket_test'
"""
from __future__ import print_function, unicode_literals

import logging

from flask_restful import Resource, abort
from flask_socketio import send
from flasgger import swag_from
from pathlib import Path

module_name = __name__.split('.')[-1]
log = logging.getLogger(module_name)

from .. import db
from .. import socketio


def setup(api, API_BASE):

    path = "/".join([API_BASE, module_name])
    log.info('Setting up "{}" and subdirectories'.format(path))
    
    api.add_resource(
        Test,
        path,
        endpoint='test',
        methods=('GET',)
    )


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class Test(Resource):

    @swag_from(str(Path(r"swagger/websocket-test.yaml")), endpoint='websocket_test')
    def get(self):
        """Return something."""
        socketio.send("you're welcome!", room='all_nodes')
        return socketio.server.manager.rooms



# -*- coding: utf-8 -*-
import logging
import base64

from marshmallow import fields
from marshmallow_sqlalchemy import (
    ModelSchema, 
    field_for, 
    ModelConverter
)
from flask import url_for, Flask
from flask_marshmallow.sqla import HyperlinkRelated
from flask_marshmallow import Schema
from marshmallow.fields import List, Integer
from werkzeug.routing import BuildError

from vantage6.server import api
from vantage6.server.util import logger_name
from vantage6.server.globals import STRING_ENCODING
from .. import ma
from .. import db

log = logging.getLogger(logger_name(__name__))


class HATEOASModelSchema(ModelSchema):
    """Convert foreign-key fields to HATEOAS specification."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # to one relationship
        setattr(self, "node",
            lambda obj: self.hateos("node", obj))
        setattr(self, "organization",
            lambda obj: self.hateos("organization", obj))
        setattr(self, "collaboration",
            lambda obj: self.hateos("collaboration", obj))
        setattr(self, "user",
            lambda obj: self.hateos("user", obj))
        setattr(self, "result",
            lambda obj: self.hateos("result", obj))
        setattr(self, "task",
            lambda obj: self.hateos("task", obj))
        
        # to many relationship
        setattr(self, "nodes", 
            lambda obj: self.hateos_list("node", obj))
        setattr(self, "organizations", 
            lambda obj: self.hateos_list("organization", obj))
        setattr(self, "collaborations", 
            lambda obj: self.hateos_list("collaboration", obj))
        setattr(self, "users", 
            lambda obj: self.hateos_list("user", obj))
        setattr(self, "results", 
            lambda obj: self.hateos_list("result", obj))
        setattr(self, "tasks", 
            lambda obj: self.hateos_list("task", obj))

        # special cases
        
    def many_hateos_from_secondary(self, first, second, obj):
        hateos_list = list()
        for elem in getattr(obj, first):
            hateos_list.append(
                self._hateos_from_related(getattr(elem, second), second)
            )
        return hateos_list
            
    def hateos_from_secondairy_model(self, first, second, obj):
        first_elem = getattr(obj, first)
        second_elem = getattr(first_elem, second)
        return self._hateos_from_related(second_elem, second)

    def hateos(self, name, obj):
        elem = getattr(obj, name)
        return self._hateos_from_related(elem, name)
    
    def hateos_list(self, name, obj):
        hateos_list = list()
        type_ = type(obj)
        for elem in getattr(obj, name+"s"):
            hateos = self._hateos_from_related(elem, name)
            hateos_list.append(hateos)
        return hateos_list
    
    def _hateos_from_related(self, elem, name):
        _id = elem.id
        endpoint = name+"_with_id"
        if not api.owns_endpoint(endpoint):
            Exception(f"Make sure {endpoint} exists!")
        
        verbs = list(api.app.url_map._rules_by_endpoint[endpoint][0].methods)
        verbs.remove("HEAD")
        verbs.remove("OPTIONS")
        url = url_for(endpoint, id=_id)
        return {"id":_id, "link":url, "methods": verbs}


# /task/{id}
class TaskSchema(HATEOASModelSchema):
    class Meta:
        model = db.Task

    complete = fields.Boolean()
    collaboration = fields.Method("collaboration")
    results = fields.Method("results")


# /task/{id}?include=result
class TaskIncludedSchema(TaskSchema):
    """Returns the TaskSchema plus the correspoding results."""
    results = fields.Nested('TaskResultSchema', many=True, exclude=['task'])


# /task/{id}/result
class TaskResultSchema(HATEOASModelSchema):
    class Meta:
        model = db.Result


class ResultSchema(HATEOASModelSchema):
    class Meta:
        model = db.Result
    
    organization = fields.Method("organization")
    task = fields.Method("task")


class ResultTaskIncludedSchema(ResultSchema):
    task = fields.Nested('TaskSchema', many=False, exclude=["results"])


class OrganizationSchema(HATEOASModelSchema):
    class Meta:
        model = db.Organization
        exclude = ('_public_key',)
    
    # convert fk to HATEOAS
    collaborations = fields.Method("collaborations")
    nodes = fields.Method("nodes")
    users = fields.Method("users")
    tasks = fields.Method("tasks")
    
    # make sure 
    public_key = fields.Function(
        lambda self: base64.b64encode(self._public_key).decode(STRING_ENCODING) if \
            self._public_key else ""
    ) 


class CollaborationSchema(HATEOASModelSchema):
    class Meta:
        model = db.Collaboration

    # convert fk to HATEOAS
    organizations = fields.Method("organizations")
    nodes = fields.Method("nodes")
    tasks = fields.Method("tasks")

# ------------------------------------------------------------------------------
class CollaborationSchemaSimple(HATEOASModelSchema):

    nodes = fields.Nested(
        'NodeSchemaSimple', 
        many=True, 
        exclude=['nodes', 'tasks', 'collaboration']
    )

    class Meta:
        table = db.Collaboration.__table__
        exclude = [
            'tasks',
            'organizations',
            # 'nodes',
        ]


# ------------------------------------------------------------------------------
class NodeSchema(HATEOASModelSchema):
    # organization = ma.HyperlinkRelated('organization_with_id')
    # collaboration = ma.HyperlinkRelated('collaboration_with_id')

    organization = fields.Method("organization")
    collaboration = fields.Method("collaboration")

    class Meta:
        model = db.Node


# ------------------------------------------------------------------------------
class NodeSchemaSimple(HATEOASModelSchema):

    # collaboration = fields.Nested(
    #     'CollaborationSchema', 
    #     many=False, 
    #     exclude=['organizations', 'nodes', 'tasks']
    # )

    organization = fields.Nested(
        'OrganizationSchema', 
        many=False, 
        exclude=[
            '_id', 
            'id', 
            'domain', 
            'address1', 
            'address2', 
            'zipcode', 
            'country',
            'nodes', 
            'collaborations', 
            'users',
            ]
    )


    class Meta:
        model = db.Node
        exclude = [
            # 'id', 
            # 'organization', 
            # 'collaboration', 
            'taskresults', 
            'api_key', 
            'type', 
        ]

# ------------------------------------------------------------------------------
class UserSchema(HATEOASModelSchema):
    
    class Meta:
        model = db.User
        exclude = ('password',)

    organization = fields.Method("organization")


# -*- coding: utf-8 -*-
"""
Resources below '/<api_base>/collaboration'
"""

from __future__ import print_function, unicode_literals

import json
import logging

from flask import request, g
from flask_restful import Resource, reqparse
from flasgger import swag_from
from http import HTTPStatus
from pathlib import Path

from vantage6.server import db
from vantage6.server.resource._schema import (
    CollaborationSchema,
    TaskSchema,
    OrganizationSchema,
)
from vantage6.server.resource import with_user_or_node, with_user, only_for

module_name = __name__.split('.')[-1]
log = logging.getLogger(module_name)


def setup(api, api_base):
    path = "/".join([api_base, module_name])
    log.info('Setting up "{}" and subdirectories'.format(path))

    api.add_resource(
        Collaboration,
        path,
        endpoint='collaboration_without_id',
        methods=('GET', 'POST')
    )
    api.add_resource(
        Collaboration,
        path + '/<int:id>',
        endpoint='collaboration_with_id',
        methods=('GET', 'PATCH', 'DELETE')
    )
    api.add_resource(
        CollaborationOrganization,
        path+'/<int:id>/organization',
        endpoint='collaboration_with_id_organization',
        methods=('GET', 'POST', 'DELETE')
    )
    api.add_resource(
        CollaborationNode,
        path+'/<int:id>/node',
        endpoint='collaboration_with_id_node',
        methods=('GET', 'POST', 'DELETE')
    )
    api.add_resource(
        CollaborationTask,
        path+'/<int:id>/task',
        endpoint='collaboration_with_id_task',
        methods=('GET', 'POST', 'DELETE')
    )


# Schemas
collaboration_schema = CollaborationSchema()
tasks_schema = TaskSchema()
org_schema = OrganizationSchema()

# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class Collaboration(Resource):

    @with_user
    @swag_from(str(Path(r"swagger/post_collaboration_without_id.yaml")), endpoint='collaboration_without_id')
    def post(self):
        """create a new collaboration"""

        parser = reqparse.RequestParser()
        parser.add_argument(
            'name',
            type=str,
            required=True,
            help="This field cannot be left blank!"
        )
        parser.add_argument(
            'organization_ids',
            type=int,
            required=True,
            action='append'
        )
        parser.add_argument(
            'encrypted',
            type=int,
            required=False
        )
        data = parser.parse_args()

        encrypted = True if data["encrypted"] == 1 else False

        collaboration = db.Collaboration(
            name=data['name'],
            organizations=[
                db.Organization.get(org_id) for org_id in data['organization_ids'] if db.Organization.get(org_id)
            ],
            encrypted=encrypted
        )

        collaboration.save()
        return collaboration_schema.dump(collaboration).data, HTTPStatus.OK

    @only_for(['user', 'node', 'container'])
    @swag_from(str(Path(r"swagger/get_collaboration_with_id.yaml")), endpoint='collaboration_with_id')
    @swag_from(str(Path(r"swagger/get_collaboration_without_id.yaml")), endpoint='collaboration_without_id')
    def get(self, id=None):
        """collaboration or list of collaborations in case no id is provided"""
        collaboration = db.Collaboration.get(id)

        # check that collaboration exists
        if not collaboration:
            return {"msg": "collaboration having id={} not found".format(id)}, HTTPStatus.NOT_FOUND  # 404

        # check if node or user have permission to view the collaboration
        # organization_ids = collaboration.get_organization_ids()
        # auth = g.user if g.user is not None else g.node
        # if auth.organization_id not in organization_ids and "admin" not in g.user.roles:
        #     log.warning("user or node from organization_id={} tries to access collaboration_id={}".format(
        #         auth.organization_id, id)
        #     )
        #     return {"msg": "you do not have permission to view this collaboration"}

        return collaboration_schema.dump(collaboration, many=not id).data, HTTPStatus.OK  # 200

    @with_user
    @swag_from(str(Path(r"swagger/patch_collaboration_with_id.yaml")), endpoint='collaboration_with_id')
    def patch(self, id):
        """update a collaboration"""
        # if "admin" not in g.user.roles:
        #     return {"msg": "only administrators can edit collaborations"}, 403  # forbidden

        collaboration = db.Collaboration.get(id)

        # check if collaboration exists
        if not collaboration:
            return {"msg": "collaboration having collaboration_id={} can not be found".format(
                id
            )}, HTTPStatus.NOT_FOUND  # 404

        # only update fields that are provided
        data = request.get_json()
        if "name" in data:
            collaboration.name = data["name"]
        if "organization_ids" in data:
            collaboration.organizations = [
                db.Organization.get(org_id) for org_id in data['organization_ids'] if db.Organization.get(org_id)
            ]
        collaboration.save()

        return collaboration_schema.dump(collaboration, many=False), HTTPStatus.OK  # 200

    @with_user
    @swag_from(str(Path(r"swagger/delete_collaboration_with_id.yaml")), endpoint='collaboration_with_id')
    def delete(self, id):
        """delete collaboration"""

        collaboration = db.Collaboration.get(id)
        if not collaboration:
            return {"msg": "collaboration id={} is not found".format(id)}, 404

        collaboration.delete()
        return {"msg": "node id={} successfully deleted".format(id)}, 200


class CollaborationOrganization(Resource):
    """Resource for /api/collaboration/<int:id>/organization."""

    @only_for(["node","user","container"])
    @swag_from(str(Path(r"swagger/get_collaboration_organization.yaml")), endpoint='collaboration_with_id_organization')
    def get(self, id):
        """Return organizations for a specific collaboration."""
        collaboration = db.Collaboration.get(id)
        if not collaboration:
            return {"msg": "collaboration having collaboration_id={} can not be found".format(
                id
            )}, HTTPStatus.NOT_FOUND

        # only users that belong to the collaboration can view collaborators
        # organization_ids = collaboration.get_organization_ids()
        # if g.user.organization_id not in organization_ids and "admin" not in g.user.roles:
        #     return {"msg": "only users that belong to this collaboration can view its organizations"}, 403

        return collaboration.organizations, HTTPStatus.OK

    @with_user
    @swag_from(str(Path(r"swagger/post_collaboration_organization.yaml")), endpoint='collaboration_with_id_organization')
    def post(self, id):
        """Add an organizations to a specific collaboration."""
        # get collaboration to which te organization should be added
        collaboration = db.Collaboration.get(id)
        if not collaboration:
            return {"msg": "collaboration having collaboration_id={} can not be found".format(
                id
            )}, HTTPStatus.NOT_FOUND

        # get the organization
        data = request.get_json()
        organization = db.Organization.get(data['id'])
        if not organization:
            return {"msg": "organization with id={} is not found"}, HTTPStatus.NOT_FOUND

        # append organization to the collaboration
        collaboration.organizations.append(organization)
        collaboration.save()
        return collaboration.organizations

    @with_user
    @swag_from(str(Path(r"swagger/delete_collaboration_organization.yaml")), endpoint='collaboration_with_id_organization')
    def delete(self, id):
        """Removes an organization from a collaboration."""
        # get collaboration from which organization should be removed
        collaboration = db.Collaboration.get(id)
        if not collaboration:
            return {"msg": "collaboration having collaboration_id={} can not be found".format(
                id
            )}, HTTPStatus.NOT_FOUND

        # get organization which should be deleted
        data = request.get_json()
        organization = db.Organization.get(data['id'])
        if not organization:
            return {"msg": "organization with id={} is not found"}, HTTPStatus.NOT_FOUND

        # delete organization and update
        collaboration.organizations.remove(organization)
        collaboration.save()
        return {"msg": "organization id={} successfully deleted from collaboration id={}".format(
            data['id'], id
        )}, HTTPStatus.OK


class CollaborationNode(Resource):
    """Resource for /api/collaboration/<int:id>/node."""

    @with_user
    @swag_from(str(Path(r"swagger/get_collaboration_node.yaml")), endpoint='collaboration_with_id_node')
    def get(self, id):
        """"Return a list of nodes that belong to the collaboration."""
        collaboration = db.Collaboration.get(id)
        if not collaboration:
            return {"msg": "collaboration having collaboration_id={} can not be found".format(
                id
            )}, HTTPStatus.NOT_FOUND

        return collaboration.nodes, HTTPStatus.OK

    @with_user
    @swag_from(str(Path(r"swagger/post_collaboration_node.yaml")), endpoint='collaboration_with_id_node')
    def post(self, id):
        """Add an node to a specific collaboration."""
        collaboration = db.Collaboration.get(id)
        if not collaboration:
            return {"msg": "collaboration having collaboration_id={} can not be found".format(
                id
            )}, HTTPStatus.NOT_FOUND

        data = request.get_json()
        node = db.Node.get(data['id'])
        if not node:
            return {"msg": "node id={} not found"}, HTTPStatus.NOT_FOUND
        if node in collaboration.nodes:
            return {"msg": "node id={} is already in collaboration id={}".format(
                data['id'], id
            )}, HTTPStatus.BAD_REQUEST

        collaboration.nodes.append(node)
        collaboration.save()
        return collaboration.nodes

    @with_user
    @swag_from(str(Path(r"swagger/delete_collaboration_node.yaml")), endpoint='collaboration_with_id_node')
    def delete(self, id):
        """Remove node from collaboration."""
        collaboration = db.Collaboration.get(id)
        if not collaboration:
            return {"msg": "collaboration having collaboration_id={} can not be found".format(
                id
            )}, 404

        data = request.get_json()
        node = db.Node.get(data['id'])
        if not node:
            return {"msg": "node id={} not found"}, 404
        if node not in collaboration.nodes:
            return {"msg": "node id={} is not part of collaboration id={}".format(data['id'], id)}, 400

        collaboration.nodes.remove(node)
        return {"msg": "node id={} removed from collaboration id={}".format(data['id'], id)}, 200


class CollaborationTask(Resource):
    """Resource for /api/collaboration/<int:id>/task."""

    @with_user_or_node
    @swag_from(str(Path(r"swagger/get_collaboration_task.yaml")), endpoint='collaboration_with_id_task')
    def get(self, id):
        """List of tasks that belong to a collaboration"""
        collaboration = db.Collaboration.get(id)
        return tasks_schema.dump(collaboration.tasks, many=True)

    @with_user
    @swag_from(str(Path(r"swagger/post_collaboration_task.yaml")), endpoint='collaboration_with_id_task')
    def post(self, id):
        """Attach new task to collaboration"""
        collaboration = db.Collaboration.get(id)
        if not collaboration:
            return {"msg": "collaboration having collaboration_id={} can not be found".format(
                id
            )}, HTTPStatus.NOT_FOUND

        data = request.get_json()
        task = db.Task(
            collaboration=collaboration,
            name=data.get('name', ''),
            description=data.get('description', ''),
            image=data.get('image', ''),
            input=data.get('input', '') if isinstance(data.get('input', ''), str) else json.dumps(data.get('input')),
        )
        task.save()

        for organization in collaboration.organizations:
            result = db.Result(
                task=task,
                organization=organization
            )
            result.save()

        return tasks_schema.dump(collaboration.tasks, many=True)

    @with_user
    @swag_from(str(Path(r"swagger/delete_collaboration_task.yaml")), endpoint='collaboration_with_id_task')
    def delete(self, id):
        """Remove task from collaboration"""
        collaboration = db.Collaboration.get(id)
        if not collaboration:
            return {"msg": "collaboration having collaboration_id={} can not be found".format(
                id
            )}, HTTPStatus.NOT_FOUND

        data = request.get_json()
        task_id = data['task_id']
        task = db.Task.get(task_id)
        if not task:
            return {"msg": "Task id={} not found".format(task_id)}, HTTPStatus.NOT_FOUND
        if task_id not in collaboration.get_task_ids():
            return {"msg": "Task id={} is not part of collaboration id={}".format(task_id, id)}, HTTPStatus.BAD_REQUEST
        task.delete()
        return {"msg": "Task id={} is removed from collaboration id={}".format(task_id, id)}, HTTPStatus.OK



# -*- coding: utf-8 -*-
"""
Resources ...
"""
import datetime
import logging
import os
import os.path
import sys

from functools import wraps

from flask import g, request
from flask_jwt_extended import get_jwt_claims, get_jwt_identity, jwt_required

from vantage6.server import db

log = logging.getLogger(__name__.split('.')[-1])

# ------------------------------------------------------------------------------
# Helper functions/decoraters ...
# ------------------------------------------------------------------------------
def only_for(types = ['user', 'node', 'container']):
    """JWT endpoint protection decorator"""
    def protection_decorator(fn):
        @wraps(fn)
        def decorator(*args, **kwargs):

            # decode JWT-token
            identity = get_jwt_identity()
            claims = get_jwt_claims()

            # check that identity has access to endpoint
            g.type = claims["type"]
            # log.debug(f"Endpoint accessed as {g.type}")

            if g.type not in types:
                msg = f"{g.type}'s are not allowed to access {request.url} ({request.method})"
                log.warning(msg)
                raise Exception(msg)

            # do some specific stuff per identity
            g.user = g.container = g.node = None

            if g.type == 'user':
                user = get_and_update_authenticatable_info(identity)
                g.user = user
                assert g.user.type == g.type

            elif g.type == 'node':
                node = get_and_update_authenticatable_info(identity)
                g.node = node
                assert g.node.type == g.type

            elif g.type == 'container':
                g.container = identity

            else:
                raise Exception(f"Unknown entity: {g.type}")

            return fn(*args, **kwargs)
        return jwt_required(decorator)
    return protection_decorator

def get_and_update_authenticatable_info(auth_id):
    """Get DB entity from ID and update info."""
    auth = db.Authenticatable.get(auth_id)
    auth.last_seen = datetime.datetime.utcnow()
    auth.ip = request.remote_addr
    auth.save()
    return auth

# create alias decorators
with_user_or_node = only_for(["user", "node"])
with_user = only_for(["user"])
with_node = only_for(["node"])
with_container = only_for(["container"])

def parse_datetime(dt=None, default=None):
    if dt:
        return datetime.datetime.strptime(dt, '%Y-%m-%dT%H:%M:%S.%f')
    return default



import logging
import os 

from sqlalchemy import Column, Integer, inspect
from sqlalchemy.orm.session import Session, sessionmaker
from sqlalchemy.ext.declarative import declarative_base, declared_attr

from sqlalchemy import create_engine
from sqlalchemy.orm.session import Session
from sqlalchemy.engine.url import make_url
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.orm.exc import NoResultFound


from vantage6.server.util import Singleton

module_name = __name__.split('.')[-1]
log = logging.getLogger(module_name)

class Database(metaclass=Singleton):

    def __init__(self):
        self.engine = None
        self.Session = None
        self.object_session = None

    def drop_all(self):
        Base.metadata.drop_all(self.engine)
        Base.metadata.create_all(bind=self.engine)


    def connect(self, URI='sqlite:////tmp/test.db', drop_all=False):
        
        self.URI = URI
        
        URL = make_url(URI)
        log.info("Initializing the database")
        log.debug("  driver:   {}".format(URL.drivername))
        log.debug("  host:     {}".format(URL.host))
        log.debug("  port:     {}".format(URL.port))
        log.debug("  database: {}".format(URL.database))
        log.debug("  username: {}".format(URL.username))

        # Make sure that the director for the file database exists.    
        if URL.host is None and URL.database:        
            os.makedirs(os.path.dirname(URL.database), exist_ok=True)

        self.engine = create_engine(URI, convert_unicode=True)
        self.Session = scoped_session(sessionmaker(autocommit=False, autoflush=False))
        self.object_session = Session.object_session

        self.Session.configure(bind=self.engine)

        Base.metadata.create_all(bind=self.engine)
        log.info("Database initialized!")

class ModelBase:
    """Declarative base that defines default attributes."""
    _hidden_attributes = []

    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()
    
    # Primary key, internal use only
    id = Column(Integer, primary_key=True)

    @classmethod
    def get(cls, id_=None, with_session=False):
        
        session = Database().Session

        if id_ is None:
            result = session.query(cls).all()
        else:
            try:
                result = session.query(cls).filter_by(id=id_).one()
            except NoResultFound:
                result = None

        if with_session:
            return result, session

        return result

    def save(self):
        if self.id is None:
            session = Database().Session
            session.add(self)
        else:
            session = Database().object_session(self)
        session.commit()

    def delete(self):
        if not self.id:
            session = Database().Session
        else:
            session = Database().object_session(self)
        session.delete(self)
        session.commit()

    def update(self, include=None, exclude=None, **kwargs):
        """Update this instance using a dictionary."""

        # Get a list of attributes available to this class.
        # This should exclude relationships!
        inst = inspect(self)
        cols = [c_attr.key for c_attr in inst.mapper.column_attrs]
        cols = set(cols)

        # Cast the list of attributes we're trying to update to a set.
        keys = set(kwargs.keys())

        # Only *keep* keys listed in `include`
        if include:
            if type(include) != type([]):
                include = [include, ]
            include = set(include)
            keys = keys & include

        # Remove any keys that are in `exclude`
        if exclude:
            if type(exclude) != type([]):
                exclude = [exclude, ]
            exclude = set(exclude)
            keys = keys - exclude

        # Keep only those keys that are proper attributes
        attrs = cols.intersection(keys)
        for attr in attrs:
            setattr(self, attr, kwargs[attr])

Base = declarative_base(cls=ModelBase)



# -*- coding: utf-8 -*-
"""
Resources below '/<api_base>/stats'
"""
import logging
import json

import flask
from flask import g, request, url_for
from flask_restful import Resource
from . import with_user_or_node, with_user, only_for
from ._schema import *
from http import HTTPStatus
from flasgger import swag_from
from pathlib import Path

import psutil

from vantage6.server import db
from vantage6.server import socketio

module_name = __name__.split('.')[-1]
log = logging.getLogger(module_name)


def setup(api, API_BASE):
    path = "/".join([API_BASE, module_name])
    log.info('Setting up "{}" and subdirectories'.format(path))

    api.add_resource(
        Stats,
        path,
        endpoint='stats',
        methods=('GET', )
    )


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class Stats(Resource):
    """Resource for /api/stats"""

    # stats_schema = StatsSchema()
    @only_for(["user", "node"])
    def get(self, id=None):
        schema = CollaborationSchemaSimple(many=True)

        return {
            'collaborations': schema.dump(db.Collaboration.get()).data
        }


    




# -*- coding: utf-8 -*-
"""
Resources below '/<api_base>/task'
"""
import logging
import json

from flask import g, request, url_for
from flask_restful import Resource
from . import with_user_or_node, with_user, only_for
from ._schema import TaskSchema, TaskIncludedSchema, TaskResultSchema
from http import HTTPStatus
from flasgger import swag_from
from pathlib import Path

from vantage6.server import db
from vantage6.server import socketio
from vantage6.server.util import (
    bytes_to_base64s,
    base64s_to_bytes,
    log_full_request,
)
from vantage6.server.globals import STRING_ENCODING

module_name = __name__.split('.')[-1]
log = logging.getLogger(module_name)


def setup(api, API_BASE):
    path = "/".join([API_BASE, module_name])
    log.info('Setting up "{}" and subdirectories'.format(path))

    api.add_resource(
        Task,
        path,
        endpoint='task_without_id',
        methods=('GET', 'POST')
    )
    api.add_resource(
        Task,
        path + '/<int:id>',
        endpoint='task_with_id',
        methods=('GET', 'DELETE')
    )
    api.add_resource(
        TaskResult,
        path + '/<int:id>/result',
        endpoint='task_result',
        methods=('GET',)
    )


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class Task(Resource):
    """Resource for /api/task"""

    task_schema = TaskSchema()
    task_result_schema = TaskIncludedSchema()

    @only_for(["user", "node", "container"])
    @swag_from(str(Path(r"swagger/get_task_with_id.yaml")), endpoint='task_with_id')
    @swag_from(str(Path(r"swagger/get_task_without_id.yaml")), endpoint='task_without_id')
    def get(self, id=None):
        task = db.Task.get(id)
        if not task:
            return {"msg": "task id={} is not found"}, HTTPStatus.NOT_FOUND

        s = self.task_result_schema if request.args.get('include') == 'results' else self.task_schema
        return s.dump(task, many=not id).data, HTTPStatus.OK


    @only_for(["user", "container"])
    @swag_from(str(Path(r"swagger/post_task_without_id.yaml")), endpoint='task_without_id')
    def post(self):
        """Create a new Task."""
        # log_full_request(request, log)

        if not request.is_json:
            return {"msg": "No JSON body found..."}, HTTPStatus.BAD_REQUEST

        data = request.get_json()

        collaboration_id = data.get('collaboration_id')
        collaboration = db.Collaboration.get(collaboration_id)

        if not collaboration:
            return {"msg": f"collaboration id={collaboration_id} not found"},\
            HTTPStatus.NOT_FOUND

        organizations_json_list = data.get('organizations')
        org_ids = [org.get("id") for org in organizations_json_list]
        db_ids = collaboration.get_organization_ids()

        # TODO what happens if user's organization does not have a node that
        # is part of a the collaboration
        # if the 'master'-flag is set to true the (master) task is executed on
        # a node in the collaboration from the organization to which the user
        # belongs. If also organization_ids are supplied, then these are ignored.
        if data.get("master", False) and g.user:
            org_ids = [g.user.organization_id]

        # Check that all organization ids are within the collaboration
        # if not all([org_id in db_ids for org_id in org_ids]):
        if not set(org_ids).issubset(db_ids):
            return {"msg": (
                f"At least one of the supplied organizations in not within "
                f"the collaboration"
            )}, HTTPStatus.BAD_REQUEST

        if g.user:
            initiator = g.user.organization
        elif g.container:
            initiator = db.Node.get(g.container["node_id"]).organization
        else:
            # FIXME: Who then? Shouldn't this raise an exception?
            initiator = None

        # Create the new task in the database
        task = db.Task(
            collaboration=collaboration,
            name=data.get('name', ''),
            description=data.get('description', ''),
            image=data.get('image', ''),
            database=data.get('database', ''),
            initiator=initiator
        )

        if g.user:
            if not self.__verify_user_permissions(g.user, task):
                return {"msg": "You lack the permission to do that!"}, HTTPStatus.UNAUTHORIZED

            # Users can only create top-level -tasks (they will not
            # have sub-tasks). Therefore, always create a new run_id.
            # FIXME: I can see no reason why users wouldn't be allowed to create
            #   subtasks? For iterative processes this would actually make a lot
            #   of sense.
            task.run_id = task.next_run_id()
            log.debug(f"New run_id {task.run_id}")

        elif g.container:
            # verify that the container has permissions to create the task
            if not self.__verify_container_permissions(g.container, task):
                return {"msg": "Container-token is not valid"}, HTTPStatus.UNAUTHORIZED

            # Tasks created by containers are always sub-tasks
            task.parent_task_id = g.container["task_id"]
            task.run_id = db.Task.get(g.container["task_id"]).run_id
            log.debug(f"Sub task from parent_task_id={task.parent_task_id}")

        # permissions ok, save to DB
        task.save()

        organizations = [db.Organization.get(id_) for id_ in org_ids]
        log.debug(f"Assigning task to {len(organizations)} nodes.")

        for organization in organizations:
            log.debug(f"Assigning task to '{organization.name}'.")

            # TODO make this cleaner...
            # get organization specific (encrypted) input
            input_ = [org.get("input") for org in organizations_json_list \
                if org.get("id")==organization.id].pop()

            if isinstance(input_, dict):
                input_ = json.dumps(input_).encode(STRING_ENCODING)

            result = db.Result(
                task=task,
                organization=organization,
                input=input_,
            )
            result.save()

        # notify nodes a new task available (only to online nodes)
        socketio.emit(
            'new_task',
            task.id,
            room=f'collaboration_{task.collaboration_id}',
            namespace='/tasks'
        )

        log.info(f"New task created for collaboration '{task.collaboration.name}'")

        if g.type == 'user':
            log.debug(f" created by: '{g.user.username}'")
        else:
            log.debug((f" created by container on node_id={g.container['node_id']}"
                       f" for (master) task_id={g.container['task_id']}"))

        log.debug(f" url: '{url_for('task_with_id', id=task.id)}'")
        log.debug(f" name: '{task.name}'")
        log.debug(f" image: '{task.image}'")

        return self.task_schema.dump(task, many=False)


    def handle_master_task(self):
        """Handle creation of a master task."""
        pass

    def handle_regular_task(self):
        """Handle creation of a regular task."""
        pass

    @staticmethod
    def __verify_user_permissions(user, task: db.Task):
        """Verify that user is permitted to create task"""

        # I have the power!
        if "root" in g.user.roles:
            return True

        # user is within organization that is part of the collaboration
        return g.user.organization_id in task.collaboration.get_organization_ids()

    @staticmethod
    def __verify_container_permissions(container, task):
        """Validates that the container is allowed to create the task."""

        # check that the image is allowed
        # if container["image"] != task.image:
        if not task.image.endswith(container["image"]):
            log.warning((f"Container from node={container['node_id']} "
                f"attempts to post a task using illegal image!?"))
            log.warning(f"  task image: {task.image}")
            log.warning(f"  container image: {container['image']}")
            return False

        # check master task is not completed yet
        if db.Task.get(container["task_id"]).complete:
            log.warning((f"Container from node={container['node_id']} attempts "
                f"to start sub-task for a completed task={container['task_id']}"))
            return False

        # check that node id is indeed part of the collaboration
        if not container["collaboration_id"] == task.collaboration.id:
            log.warning((f"Container attempts to create a task outside "
                f"collaboration_id={container['collaboration_id']} in "
                f"collaboration_id={task.collaboration_id}!"))
            return False

        return True

    @only_for(['user'])
    @swag_from(str(Path(r"swagger/delete_task_with_id.yaml")), endpoint='task_with_id')
    def delete(self, id):
        """Deletes a task"""
        # TODO we might want to delete the corresponding results also?

        task = db.Task.get(id)
        if not task:
            return {"msg": "task id={} not found".format(id)}, HTTPStatus.NOT_FOUND

        task.delete()
        return {"msg": "task id={} successfully deleted".format(id)}, HTTPStatus.OK


class TaskResult(Resource):
    """Resource for /api/task/<int:id>/result"""

    task_result_schema = TaskResultSchema()

    @only_for(['user', 'container'])
    @swag_from(str(Path(r"swagger/get_task_result.yaml")), endpoint='task_result')
    def get(self, id):
        """Return results for task."""
        task = db.Task.get(id)
        if not task:
            return {"msg": "task id={} not found".format(id)}, HTTPStatus.NOT_FOUND

        return self.task_result_schema.dump(task.results, many=True).data, HTTPStatus.OK




# -*- coding: utf-8 -*-
"""
Resources below '/<api_base>/node'
"""

import logging
import uuid

import json

from flask import g, request
from flask_restful import Resource, reqparse
from http import HTTPStatus
from flasgger.utils import swag_from
from . import with_user_or_node, with_user
from pathlib import Path


from vantage6.server.model.base import Database

from ._schema import *

module_name = __name__.split('.')[-1]
log = logging.getLogger(module_name)


def setup(api, api_base):
    path = "/".join([api_base, module_name])
    log.info('Setting up "{}" and subdirectories'.format(path))

    api.add_resource(
        Node,
        path,
        endpoint='node_without_id',
        methods=('GET', 'POST')
    )
    api.add_resource(
        Node,
        path + '/<int:id>',
        endpoint='node_with_id',
        methods=('GET', 'DELETE', 'PATCH')
    )
    api.add_resource(
        NodeTasks,
        path + '/<int:id>/task',
        endpoint='node_tasks',
        methods=('GET', )
    )


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------

class Node(Resource):

    # Schemas
    node_schema = NodeSchema()

    @with_user_or_node
    @swag_from(str(Path(r"swagger/get_node_with_id.yaml")),
        endpoint='node_with_id')
    @swag_from(str(Path(r"swagger/get_node_without_id.yaml")),
        endpoint='node_without_id')
    def get(self, id=None):
        results = db.Node.get(id)
        user_or_node = g.user or g.node

        is_root = False

        # Only users can be root, not containers.
        if g.user:
            is_root = g.user.username == 'root'

        if id:
            if not results:
                msg = {"msg": f"Couldn't find node {id}"}
                return msg, HTTPStatus.NOT_FOUND

            if is_root:
                # Let's not make a fuss ...
                return self.node_schema.dump(results, many=False)

            if (results.organization_id != user_or_node.organization_id) \
                    and 'admin' not in g.user.roles:
                msg = {"msg": "you are not allowed to see this node"}
                return msg, HTTPStatus.FORBIDDEN  # 403

        else:
            if not is_root:
                # only the results of the user's organization are returned
                org_id = g.user.organization_id
                results = [n for n in results if n.organization_id == org_id]

        return self.node_schema.dump(results, many=not id).data, HTTPStatus.OK  # 200

    @with_user
    @swag_from(str(Path(r"swagger/post_node_without_node_id.yaml")), endpoint='node_without_id')
    def post(self):

        parser = reqparse.RequestParser()
        parser.add_argument(
            "collaboration_id",
            type=int,
            required=True,
            help="This field cannot be left blank!"
        )
        data = parser.parse_args()

        collaboration = db.Collaboration.get(data["collaboration_id"])

        # check that the collaboration exists
        if not collaboration:
            return {"msg": "collaboration_id '{}' does not exist".format(
                data["collaboration_id"]
            )}, HTTPStatus.NOT_FOUND  # 404

        # new api-key which node can use to authenticate
        api_key = str(uuid.uuid1())

        # store the new node
        # TODO an admin does not have to belong to an organization?
        # TODO we need to check that the organization belongs to the collaboration
        organization = g.user.organization
        node = db.Node(
            name="{} - {} Node".format(organization.name, collaboration.name),
            collaboration=collaboration,
            organization=organization,
            api_key=api_key
        )
        node.save()

        return self.node_schema.dump(node).data, HTTPStatus.CREATED  # 201

    @with_user
    @swag_from(str(Path(r"swagger/delete_node_with_id.yaml")), endpoint='node_with_id')
    def delete(self, id):
        """delete node account"""
        node = db.Node.get(id)

        if not node:
            return {"msg": "node with id={} not found".format(id)}, HTTPStatus.NOT_FOUND  # 404

        if node.organization_id != g.user.organization_id and g.user.roles != 'admin':
            return {"msg": "you are not allowed to delete this node"}, HTTPStatus.FORBIDDEN  # 403

        node.delete()

        return {"msg": "successfully deleted node id={}".format(id)}, HTTPStatus.OK  # 200

    @with_user_or_node
    @swag_from(str(Path(r"swagger/patch_node_with_id.yaml")), endpoint='node_with_id')
    def patch(self, id):
        """update existing node"""
        node = db.Node.get(id)

        # do not create new nodes here
        if not node:
            return {"msg": "Use POST to create a new node"}, HTTPStatus.FORBIDDEN  # 403

        data = request.get_json()
        if 'state' in data:
            data['state'] = json.dumps(data['state'])

        if g.node:
            if g.node.id == node.id:
                log.debug("Hey! It's me! I got this!")
                node.update(include=['status', 'state'], **data)
            else:
                log.debug("This doesn't seem right! You don't look like me!?")
                return {"msg": "you are not allowed to edit this node"}, HTTPStatus.FORBIDDEN  # 403

        if g.user:
            is_root = g.user.roles == 'root'

            if is_root:
                # root can do everything ... he's really cool
                log.debug('root is making changes!')
                node.update(**data)
                node.save()

            else:
                # Ok, so we're not root ...
                incorrect_organization = node.organization_id != g.user.organization_id
                incorrect_role = g.user.roles != 'admin'

                if (incorrect_organization or incorrect_role):
                    return {"msg": "you are not allowed to edit this node"}, HTTPStatus.FORBIDDEN  # 403

                # We've established you're an admin for your organisation. Feel free
                # to make some changes to the api_key, name or status.
                allowed_attrs = ['name', 'api_key', 'status', 'state']
                node.update(include=allowed_attrs, **data)
                node.save()

        return self.node_schema.dump(node)  # 200


class NodeTasks(Resource):
    """Resource for /api/node/<int:id>/task.
    returns task(s) belonging to a specific node

       Resource for /api/node/<int:id>/task/<int:id>
    returns


    TODO do we need the second usage? we can retrieve tasks by the endpoint /api/task
    TODO if we do want to keep this, we need to make sure the user only sees task that belong to this node
    TODO also the user can only see nodes which belong to their organization
    """

    task_result_schema = TaskResultSchema()

    @with_user_or_node
    @swag_from(str(Path(r"swagger/get_node_tasks.yaml")), endpoint='node_tasks')
    def get(self, id):
        """Return a list of tasks for a node or a single task <task_result_id> belonging t.

        If the query parameter 'state' equals 'open' the list is
        filtered to return only tasks without result.
        """
        # get tasks that belong to node <id>
        node = db.Node.get(id)
        if not node:
            return {"msg": "node with id={} not found".format(id)}, \
                HTTPStatus.NOT_FOUND  # 404

        # select tasks from organization that are within the collaboration
        # of the node
        results = []
        for result in node.organization.results:

            if node.collaboration == result.task.collaboration:
                # result belongs to node
                if request.args.get('state') == 'open':
                    if result.complete:
                        results.append(result)
                else:
                    results.append(result)

        return self.task_result_schema.dump(results, many=True), \
            HTTPStatus.OK  # 200


# -*- coding: utf-8 -*-
import os, sys
import importlib
import flask_socketio
import datetime
import logging
import uuid

TERMINAL_AVAILABLE = True
try:
    # Stuff needed for running shell in a browser
    import pty
    import select
    import subprocess
    import struct
    import fcntl
    import termios
except:
    TERMINAL_AVAILABLE = False

from flask import Flask, Response, request, render_template, make_response, g, session
from flask_restful import Resource, Api, fields
from flask_cors import CORS
from flask_jwt_extended import JWTManager, get_jwt_identity, get_jwt_claims, get_raw_jwt, jwt_required, jwt_optional, verify_jwt_in_request

from flask_marshmallow import Marshmallow
from flask_socketio import SocketIO, emit, send,join_room, leave_room

from flasgger import Swagger

import json

from ._version import version_info, __version__
from vantage6.server import db

from vantage6.server import util
from vantage6.server.globals import APPNAME
from vantage6.server.websockets import DefaultSocketNamespace
from .resource.swagger import swagger_template

module_name = __name__.split('.')[-1]
log = logging.getLogger(module_name)


# ------------------------------------------------------------------------------
# Initialize Flask
# ------------------------------------------------------------------------------
RESOURCES_INITIALIZED = False
API_BASE = '/api'
WEB_BASE = '/app'

# Create Flask app
ROOT_PATH = os.path.dirname(__file__)
app = Flask(APPNAME, root_path=ROOT_PATH)

app.config['JWT_AUTH_URL_RULE'] ='/api/token'

# False means refresh tokens never expire
app.config['JWT_REFRESH_TOKEN_EXPIRES'] = False

# Open Api Specification (f.k.a. swagger)
app.config['SWAGGER'] = {
    'title': APPNAME + ' API',
    'uiversion': 3,
    'openapi': '3.0.0',
}
swagger = Swagger(app, template=swagger_template)

# Enable cross-origin resource sharing
CORS(app)


# ------------------------------------------------------------------------------
# Api - REST JSON-rpc
# ------------------------------------------------------------------------------
api = Api(app)

@api.representation('application/json')
def output_json(data, code, headers=None):

    if isinstance(data, db.Base):
        data = db.jsonable(data)
        # log.debug("json-proofed")
    elif isinstance(data, list) and len(data) and isinstance(data[0], db.Base):
        data = db.jsonable(data)
        # log.debug("json-list-proofed")
    # log.debug(f"finished preparing {data}, lets send")

    resp = make_response(json.dumps(data), code)
    resp.headers.extend(headers or {})
    return resp


# ------------------------------------------------------------------------------
# Setup SQLAlchemy and Marshmallow for marshalling/serializing
# ------------------------------------------------------------------------------
ma = Marshmallow(app)

# ------------------------------------------------------------------------------
# Setup the Flask-JWT-Extended extension (JWT: JSON Web Token)
# ------------------------------------------------------------------------------
jwt = JWTManager(app)

@jwt.user_claims_loader
def user_claims_loader(identity):
    roles = []
    if isinstance(identity, db.User):
        type_ = 'user'
        roles = identity.roles.split(',')
    elif isinstance(identity, db.Node):
        type_ = 'node'
    elif isinstance(identity, dict):
        type_ = 'container'
    else:
        log.error(f"could not create claims from {str(identity)}")

    claims = {
        'type': type_,
        'roles': roles,
    }

    return claims

@jwt.user_identity_loader
def user_identity_loader(identity):

    if isinstance(identity, db.Authenticatable):
        return identity.id
    if isinstance(identity, dict):
        return identity

    log.error(f"Could not create a JSON serializable identity \
                from '{str(identity)}'")

@jwt.user_loader_callback_loader
def user_loader_callback(identity):
    if isinstance(identity, int):
        return db.Authenticatable.get(identity)
    else:
        return identity


# ------------------------------------------------------------------------------
# Setup flask-socketio
# ------------------------------------------------------------------------------
try:
    socketio = SocketIO(app, async_mode='gevent_uwsgi')
except:
    socketio = SocketIO(app)

# socketio.on_namespace(DefaultSocketNamespace("/"))
socketio.on_namespace(DefaultSocketNamespace("/tasks"))


def start_interpreter():
    # create child process attached to a pty we can read from and write to
    if TERMINAL_AVAILABLE:
        env = app.config['environment']
        cmd = ['ipython', '-m', 'vantage6.server.shell', '-i', '--', env]

        log.debug("opening pty")
        master_fd, slave_fd = pty.openpty()

        log.debug("starting process")
        child = subprocess.Popen(
            cmd,
            stdin=slave_fd,
            stdout=slave_fd,
            stderr=slave_fd
        )

        log.debug("adding process details to session")
        session.child = child
        session.fd = master_fd
        session.master_fd = master_fd
        session.slave_fd = slave_fd

        log.debug("setting window size")
        set_winsize(master_fd, 50, 50)

        log.debug("starting background task")
        socketio.start_background_task(
            read_and_forward_pty_output,
            fd=master_fd,
            sid=request.sid,
            child=child,
        )
        log.debug("ipython terminal backend started")
    else:
        log.debug("ipython terminal not available")


def assert_running_interpreter(start_if_required=False, child=None):
    # log.debug(f"assert_running_interpreter(start_if_required={start_if_required})")

    try:
        child = child or session.child

        if child.poll() is None:
            # log.debug("interpreter already running!")
            return True
    except (AttributeError, TypeError):
        pass


    if start_if_required:
        # log.debug("starting interpreter")
        start_interpreter()
        return True

    return False

def set_winsize(fd, row, col, xpix=0, ypix=0):
    winsize = struct.pack("HHHH", row, col, xpix, ypix)
    fcntl.ioctl(fd, termios.TIOCSWINSZ, winsize)


def read_and_forward_pty_output(fd, sid, child):
    max_read_bytes = 1024 * 20

    # fd = session.fd
    timeout_sec = 0.01

    while assert_running_interpreter(child=child):
        socketio.sleep(timeout_sec)

        (rs, ws, es) = select.select([fd], [], [], timeout_sec)

        for r in rs:
            output = os.read(r, max_read_bytes).decode()
            socketio.emit(
                "pty-output",
                {"output": output},
                namespace="/pty",
                room=sid,
            )


@socketio.on("connect", namespace="/pty")
def connect_pty():
    """new client connected"""
    environment = app.config['environment']

    log.debug('-' * 80)
    log.debug('connect /pty')
    log.debug(f'environment: {environment}')
    log.debug('-' * 80)

    try:
        verify_jwt_in_request()
    except Exception as e:
        log.error("Could not connect client! No or Invalid JWT token?")
        log.info(list(request.headers.keys()))
        log.exception(e)
    else:
        # At this point we're sure that the user/client/whatever
        # checks out
        user_or_node_id = get_jwt_identity()
        auth = db.Authenticatable.get(user_or_node_id)

        if not isinstance(auth, db.User):
            log.error("Sorry, but only users can use this websocket")
            return False

        if auth.username != 'root':
            log.error("Only root can connect to the admin channel")
            log.error(f"You're trying to connect as '{auth.username}'")
            return False

        # define socket-session variables.
        session.type = auth.type
        session.name = auth.username if session.type == 'user' else auth.name
        log.info(f'Client identified as <{session.type}>: <{session.name}>')

        assert_running_interpreter(start_if_required=True)
        return True

    return False

@socketio.on("disconnect", namespace="/pty")
def disconnect_pty():
    print(f'Client {request.sid} disconnected')
    # app.config['socket_connections'].remove(request.sid)
    # session["child"].kill()
    try:
        session.child.kill()
    except Exception as e:
        log.error("Could not kill interpreter backend!?")
        log.exception(e)



@socketio.on("pty-input", namespace="/pty")
def pty_input(data):
    """write to the child pty. The pty sees this as if you are typing in a real
    terminal.
    """
    if assert_running_interpreter():
        raw_data = data["input"].encode()
        # if raw_data == '':
        #     log.error('empty string!?')
        # else:
        #     log.info(f"input: '{raw_data}'")

        # os.write(app.config["fd"], raw_data)
        os.write(session.fd, raw_data)


@socketio.on("resize", namespace="/pty")
def resize(data):
    if assert_running_interpreter():
        set_winsize(session.fd, data["rows"], data["cols"])


@socketio.on("connect", namespace="/admin")
def connect_admin():
    environment = app.config['environment']

    print('-' * 80)
    print('connect /admin')
    print(f'environment: {environment}')
    print('-' * 80)

    try:
        verify_jwt_in_request()
    except Exception as e:
        log.error("Could not connect client! No or Invalid JWT token?")
        log.info(list(request.headers.keys()))
        log.exception(e)
    else:
        # At this point we're sure that the user/client/whatever
        # checks out
        user_or_node_id = get_jwt_identity()
        auth = db.Authenticatable.get(user_or_node_id)

        if not isinstance(auth, db.User):
            log.error("Sorry, but only users can use this websocket")
            return False

        if auth.username != 'root':
            log.error("Only root can connect to the admin channel")
            log.error(f"You're trying to connect as '{auth.username}'")
            return False

        # define socket-session variables.
        session.type = auth.type
        session.name = auth.username if session.type == 'user' else auth.name
        log.info(f'Client identified as <{session.type}>: <{session.name}>')

        print(f'connecting {request.sid}')
        return True

    return False


@socketio.on("echo", namespace="/")
def echo(data):
    print("Hello!")
    socketio.emit('echo', data, namespace='/')


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
def load_resources(api, API_BASE, resources):
    """Import the modules containing Resources."""
    for name in resources:
        module = importlib.import_module('vantage6.server.resource.' + name)
        module.setup(api, API_BASE)


# ------------------------------------------------------------------------------
# Http routes
# ------------------------------------------------------------------------------
@app.route(WEB_BASE+'/', defaults={'path': ''})
# @app.route(WEB_BASE+'/<path:path>')
def index(path):
    return """
    <html>
        <head>
        <script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/socket.io/2.1.1/socket.io.dev.js"></script>
        <script type="text/javascript" charset="utf-8">
            setTimeout(function() {
                console.log('running javascript');
                socket = io.connect('http://' + document.domain + ':' + location.port, { transports: ['websocket']});

                socket.on('message', function(msg){
                    console.log(msg);
                });

                socket.on('connect', function() {
                    console.log('emitting!');
                    socket.emit("my event", {data: "I'm connected!"});
                });},
                2000
            );
        </script>
        </head>
        <body>
            <h1>Hi there!!</h2>
        </body>
    </html>
"""


# ------------------------------------------------------------------------------
# init & run
# ------------------------------------------------------------------------------
# def init(environment=None, init_resources_=False):
#     """Initialize the server using a site-wide ServerContext."""
#     logging.getLogger("urllib3").setLevel(logging.WARNING)

#     if environment is None:
#         if 'environment' in os.environ:
#             environment = os.environ['environment']
#         else:
#             environment = 'prod'

#     # Load configuration and initialize logging system
#     ctx = util.ServerContext(APPNAME, 'default')
#     ctx.init(ctx.config_file, environment)

#     if init_resources_:
#         init_resources(ctx)

#     uri = ctx.get_database_location()
#     print('-' * 80)
#     print(uri)
#     print('-' * 80)
#     db.init(uri)

class WebSocketLoggingHandler(logging.Handler):

    def emit(self, record):
        entry = self.format(record)
        # print(f'emitting to websocket: {entry}')

        socketio.emit('append-log', entry, namespace='/admin')


def init_resources(ctx):
    # Load resources
    global RESOURCES_INITIALIZED

    if RESOURCES_INITIALIZED:
        return

    api_base = ctx.config['api_path']

    resources = [
            'node',
            'collaboration',
            'organization',
            'task',
            'result',
            'token',
            'user',
            'version',
            # 'websocket_test',
            'stats',
    ]

    # Load resources
    load_resources(api, api_base, resources)

    # Make sure we do this only once
    RESOURCES_INITIALIZED = True


def run(ctx, *args, **kwargs):
    """Run the server.

        Note that this method is never called when the server is instantiated
        through the Web Server Gateway Interface (WSGI)!
    """
    # Prevent logging from urllib3
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("socketIO-client").setLevel(logging.WARNING)

    wslh = WebSocketLoggingHandler()
    wslh.setLevel(logging.DEBUG)
    logging.getLogger().addHandler(wslh)

    environment = ctx.config.get('type')
    app.config['environment'] = environment

    app.config['JWT_SECRET_KEY'] = ctx.config.get('jwt_secret_key', str(uuid.uuid1()))

    # Default expiration time
    app.config['JWT_ACCESS_TOKEN_EXPIRES'] = datetime.timedelta(hours=6)
    # Set an extra long expiration time on access tokens for testing

    if environment == 'test':
        log.warning("Setting 'JWT_ACCESS_TOKEN_EXPIRES' to one day!")
        app.config['JWT_ACCESS_TOKEN_EXPIRES'] = datetime.timedelta(days=1)

    # check if root-user exists
    try:
        db.User.getByUsername("root")
    except Exception:
        log.warn("No root user found! Is this the first run?")
        log.warn("Creating root: username=root, password=root")
        user = db.User(username="root", roles="root")
        user.set_password("root")
        user.save()

    # set all nodes to offline
    nodes, session = db.Node.get(with_session=True)
    for node in nodes:
        node.status = 'offline'
    session.commit()

    kwargs.setdefault('log_output', True)
    socketio.run(app, *args, **kwargs)




# -*- coding: utf-8 -*-
"""
Resources below '/<api_base>/organization'
"""
from __future__ import print_function, unicode_literals

import logging
import base64

from flask import request
from flask_restful import Resource
from flasgger import swag_from
from http import HTTPStatus
from pathlib import Path

from vantage6.server.resource import with_user_or_node, with_user, only_for
from ._schema import *


module_name = __name__.split('.')[-1]
log = logging.getLogger(module_name)


def setup(api, API_BASE):

    path = "/".join([API_BASE, module_name])
    log.info('Setting up "{}" and subdirectories'.format(path))

    api.add_resource(
        Organization,
        path,
        endpoint='organization_without_id',
        methods=('GET', 'POST')
     )
    api.add_resource(
        Organization,
        path + '/<int:id>',
        endpoint='organization_with_id',
        methods=('GET', 'PATCH')
    )
    api.add_resource(
        OrganizationCollaboration,
        path + '/<int:id>/collaboration',
        endpoint='organization_collaboration',
        methods=('GET',)
    )
    api.add_resource(
        OrganizationNode,
        path + '/<int:id>/node',
        endpoint='organization_node',
        methods=('GET',)
    )


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class Organization(Resource):

    org_schema = OrganizationSchema()

    @only_for(["user", "node", "container"])
    @swag_from(str(Path(r"swagger/get_organization_with_id.yaml")), 
        endpoint='organization_with_id')
    @swag_from(str(Path(r"swagger/get_organization_without_id.yaml")), 
        endpoint='organization_without_id')
    def get(self, id=None):
        organization = db.Organization.get(id)
        if not organization:
            return {"msg": f"organization id={id} not found"}, \
                HTTPStatus.NOT_FOUND

        return self.org_schema.dump(organization, many=not id).data, \
            HTTPStatus.OK

    @only_for(["user"])
    @swag_from(str(Path(r"swagger/post_organization_without_id.yaml")), 
        endpoint='organization_without_id')
    def post(self):
        """Create a new organization."""

        data = request.get_json()
        organization = db.Organization(
            name=data.get('name', ''),
            address1=data.get('address1', ''),
            address2=data.get('address2' ''),
            zipcode=data.get('zipcode', ''),
            country=data.get('country', ''),
            public_key=data.get('public_key', '')
        )
        organization.save()

        return self.org_schema.dump(
            organization, many=False
        ).data, HTTPStatus.CREATED

    @only_for(["user", "node"])
    def patch(self, id):
        """Update organization."""
        
        organization = db.Organization.get(id)
        if not organization:
            return {"msg": f"Organization with id {id} not found"}, \
                HTTPStatus.NOT_FOUND
        
        data = request.get_json()
        fields = ["name", "address1", "address2", "zipcode", "country", \
            "public_key"]
        for field in fields:
            if data.get(field):
                setattr(organization, field, data.get(field))
        
        organization.save()
        return organization, HTTPStatus.OK

class OrganizationCollaboration(Resource):
    """Collaborations for a specific organization."""

    col_schema = CollaborationSchema()

    @only_for(["user", "node"])
    @swag_from(str(Path(r"swagger/get_organization_collaboration.yaml")), endpoint='organization_collaboration')
    def get(self, id):
        organization = db.Organization.get(id)
        if not organization:
            return {"msg": "organization id={} not found".format(id)}, HTTPStatus.NOT_FOUND

        return self.col_schema.dump(organization.collaborations, many=True).data, HTTPStatus.OK

    # @with_user
    # def post(self, id):
    #     organization = db.Organization.get(id)
    #     if not organization:
    #         return {"msg": "organization id={} not found".format(id)}, HTTPStatus.NOT_FOUND
    #
    #     data = request.get_json()
    #     collaboration = db.Collaboration.get(data['id'])
    #     if not collaboration:
    #         return {"msg": "collaboration id={} is not found".format(data['id'])}, HTTPStatus.NOT_FOUND
    #
    #     organization.collaborations.append(collaboration)
    #     organization.save()
    #     return self.col_schema.dump(organization.collaborations, many=True).data, HTTPStatus.OK


class OrganizationNode(Resource):
    """Resource for /api/organization/<int:id>/node."""

    nod_schema = NodeSchema()

    @with_user_or_node
    @swag_from(str(Path(r"swagger/get_organization_node.yaml")), endpoint='organization_node')
    def get(self, id):
        """Return a list of Nodes."""
        organization = db.Organization.get(id)
        if not organization:
            return {"msg": "organization id={} not found".format(id)}, HTTPStatus.NOT_FOUND

        return self.nod_schema.dump(organization.nodes, many=True).data, HTTPStatus.OK

    # @with_user
    # def post(self, id):
    #     """Create new node"""
    #     organization = db.Organization.get(id)
    #     if not organization:
    #         return {"msg": "organization id={} not found".format(id)}, HTTPStatus.NOT_FOUND
    #
    #     data = request.get_json()
    #
    #     db.Node(
    #         name="{} - {} Node".format(organization.name, collaboration.name),
    #         collaboration=collaboration,
    #         organization=organization,
    #         api_key=api_key
    #     )
    #     node.save()
    #
    #     return node





from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.ext.hybrid import hybrid_property

from .base import Database
from .authenticable import Authenticatable

class Node(Authenticatable):
    """Application that executes Tasks."""
    _hidden_attributes = ['api_key']

    id = Column(Integer, ForeignKey('authenticatable.id'), primary_key=True)
    
    # fields
    name = Column(String)
    api_key = Column(String)
    collaboration_id = Column(Integer, ForeignKey("collaboration.id"))
    organization_id = Column(Integer, ForeignKey("organization.id"))
    
    # relationships
    collaboration = relationship("Collaboration", back_populates='nodes')
    organization = relationship("Organization", back_populates='nodes')

    # the type specification in Authenticatable 
    __mapper_args__ = {
        'polymorphic_identity': 'node',
    }

    @classmethod
    def get_by_api_key(cls, api_key):
        """returns Node based on the provided API key.
        
        Returns None if no Node is associated with api_key.
        """
        session = Database().Session

        try:
            return session.query(cls).filter_by(api_key=api_key).one()
        except NoResultFound:
            return None

    def __repr__(self):
        return ("<Node "
            f"name: {self.name}, "
            f"organization: {self.organization.name}, "
            f"collaboration: {self.collaboration.name}, "
            f"api_key: {self.api_key} "
        ">")


# -*- coding: utf-8 -*-
import types
import yaml
import unittest
import doctest
import logging
import json

from flask import Flask, Response as BaseResponse, json
from flask.testing import FlaskClient
from werkzeug.utils import cached_property

from vantage6.common.globals import APPNAME #, VERSION
from vantage6.server.globals import PACAKAGE_FOLDER
from vantage6 import server
from vantage6.server import (
    util,
    db,
    context
)
from vantage6.server.controller.fixture import load
from vantage6.server.model.base import Database

log = logging.getLogger(__name__.split('.')[-1])


class Response(BaseResponse):
    @cached_property
    def json(self):
        return json.loads(self.data)


class TestNode(FlaskClient):
    def open(self, *args, **kwargs):
        if 'json' in kwargs:
            kwargs['data'] = json.dumps(kwargs.pop('json'))
            kwargs['content_type'] = 'application/json'
        return super().open(*args, **kwargs)


class TestResources(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.loglevels = {}

        for lib in ['fixtures', 'db']:
            l = logging.getLogger(lib)
            cls.loglevels[lib] = l.level
            l.setLevel(logging.WARNING)

    @classmethod
    def tearDownClass(cls):
        for lib in cls.loglevels:
            l = logging.getLogger(lib)
            l.setLevel(cls.loglevels[lib])

    def setUp(self):
        """Called immediately before running a test method."""
        Database().connect("sqlite://")
        file_ = str(PACAKAGE_FOLDER / APPNAME / "server" / "_data" /
                    "unittest_fixtures.yaml")
        with open(file_) as f:
            self.entities = yaml.safe_load(f.read())
        load(self.entities, drop_all=True)

        server.app.testing = True
        server.app.response_class = Response
        server.app.test_client_class = TestNode
        server.app.secret_key = "test-secret"


        ctx = context.TestContext.from_external_config_file(
            "unittest_config.yaml")

        server.init_resources(ctx)

        self.app = server.app.test_client()

        self.credentials = {
            'root': {
                'username': 'root',
                'password': 'root'
            },
            'admin':{
                'username': 'frank@iknl.nl',
                'password': 'password'
            },
            'user':{
                'username': 'melle@iknl.nl',
                'password': 'password'
            }
        }

    def login(self, type_='root'):
        tokens = self.app.post(
            '/api/token/user',
            json=self.credentials[type_]
        ).json
        headers = {
            'Authorization': 'Bearer {}'.format(tokens['access_token'])
        }
        return headers

    def test_version(self):
        rv = self.app.get('/api/version')
        r = json.loads(rv.data)
        self.assertIn('version', r)
        self.assertEqual(r['version'], vantage6.server.__version__)

    def test_token_different_users(self):
        for type_ in ["root", "admin", "user"]:
            tokens = self.app.post('/api/token/user',
                json=self.credentials[type_]).json
            self.assertIn('access_token', tokens)
            self.assertIn('refresh_token', tokens)

    def test_organization(self):
        headers = self.login()

        # First retrieve a list of all organizations
        orgs = self.app.get('/api/organization', headers=headers).json
        self.assertEqual(len(orgs), 3)

        attrs = [
            'id',
            'name',
            'domain',
            'name',
            'address1',
            'address2',
            'zipcode',
            'country',
        ]

        org = orgs[0]
        for attr in attrs:
            self.assertIn(attr, org)

        # Retrieve a single organization
        url = '/api/organization/{}'.format(org['id'])
        org = self.app.get(url, headers=headers).json
        self.assertEqual(org['id'], orgs[0]['id'])
        self.assertEqual(org['name'], orgs[0]['name'])

        # Create a new organization
        org_details = {
            'name': 'Umbrella Corporation',
            'address1': 'Resident Evil Pike'
        }

        org = self.app.post(
            '/api/organization',
            json=org_details,
            headers=headers
        ).json

        # for attr in attrs:
        #     self.assertIn(attr, org)

        # self.assertGreater(org['id'], 0)

        orgs = self.app.get('/api/organization', headers=headers).json
        # self.assertEqual(len(orgs), 4)

    def test_collaboration(self):
        headers = self.login()

        collaborations = self.app.get(
            '/api/collaboration', headers=headers
        ).json
        self.assertEqual(len(collaborations), 3)

    def test_node_without_id(self):

        # GET
        headers = self.login("root")
        nodes = self.app.get("/api/node", headers=headers).json
        expected_fields = [
            'name',
            'api_key',
            'collaboration',
            'organization',
            'status',
            'id',
            'type',
            'last_seen',
            'ip'
        ]
        for node in nodes:
            for key in expected_fields:
                self.assertIn(key, node)

        headers = self.login("user")
        nodes = self.app.get("/api/node", headers=headers).json
        self.assertIsNotNone(nodes)

        # POST
        # unknown collaboration id should fail
        response = self.app.post("/api/node", headers=headers, json={
            "collaboration_id": 99999
        })
        response_json = response.json
        self.assertIn("msg", response_json)
        self.assertEqual(response.status_code, 404) # NOT FOUND

        # succesfully create a node
        response = self.app.post("/api/node", headers=headers, json={
            "collaboration_id": 1
        })
        response_json = response.json
        self.assertEqual(response.status_code, 201) # CREATED

    def test_node_with_id(self):

        # root user can access all nodes
        headers = self.login("root")
        node = self.app.get("/api/node/8", headers=headers).json
        expected_fields = [
            'name',
            'api_key',
            'collaboration',
            'organization',
            'status',
            'id',
            'type',
            'last_seen',
            'ip'
        ]
        for key in expected_fields:
            self.assertIn(key, node)

        # user cannot access all
        headers = self.login("user")
        node = self.app.get("/api/node/8", headers=headers)
        self.assertEqual(node.status_code, 403)

        # some nodes just don't exist
        node = self.app.get("/api/node/9999", headers=headers)
        self.assertEqual(node.status_code, 404)

    def test_node_tasks(self):
        headers = self.login("root")

        # Non existing node
        task = self.app.get("/api/node/9999/task", headers=headers)
        self.assertEqual(task.status_code, 404)

        task = self.app.get("/api/node/7/task", headers=headers)
        self.assertEqual(task.status_code, 200)

        task = self.app.get("/api/node/7/task?state=open", headers=headers)
        self.assertEqual(task.status_code, 200)

    def test_result_with_id(self):
        headers = self.login("root")
        result = self.app.get("/api/result/1", headers=headers)
        self.assertEqual(result.status_code, 200)

        result = self.app.get("/api/result/1?include=task", headers=headers)
        self.assertEqual(result.status_code, 200)

    def test_result_without_id(self):
        headers = self.login("root")
        result = self.app.get("/api/result", headers=headers)
        self.assertEqual(result.status_code, 200)

        result = self.app.get("/api/result?state=open&&node_id=1",
            headers=headers)
        self.assertEqual(result.status_code, 200)

        result = self.app.get("/api/result?task_id=1", headers=headers)
        self.assertEqual(result.status_code, 200)

        result = self.app.get("/api/result?task_id=1&&node_id=1", headers=headers)
        self.assertEqual(result.status_code, 200)

    def test_stats(self):
        headers = self.login("root")
        result = self.app.get("/api/result", headers=headers)
        self.assertEqual(result.status_code, 200)

    def test_task_with_id(self):
        headers = self.login("root")
        result = self.app.get("/api/task/1", headers=headers)
        self.assertEqual(result.status_code, 200)

    def test_task_witout_id(self):
        headers = self.login("root")
        result = self.app.get("/api/task", headers=headers)
        self.assertEqual(result.status_code, 200)

    def test_task_including_results(self):
        headers = self.login("root")
        result = self.app.get("/api/task?include=results", headers=headers)
        self.assertEqual(result.status_code, 200)

    def test_task_unknown(self):
        headers = self.login("root")
        result = self.app.get("/api/task/9999", headers=headers)
        self.assertEqual(result.status_code, 404)

    def test_user_with_id(self):
        headers = self.login("admin")
        result = self.app.get("/api/user/1", headers=headers)
        self.assertEqual(result.status_code, 200)
        user = result.json

        expected_fields = [
            "username",
            "firstname",
            "lastname",
            "roles"
        ]
        for field in expected_fields:
            self.assertIn(field, user)

    def test_user_unknown(self):
        headers = self.login("admin")
        result = self.app.get("/api/user/9999", headers=headers)
        self.assertEqual(result.status_code, 404)

    def test_user_without_id(self):
        for role in ["user", "admin", "root"]:
            headers = self.login(role)
            result = self.app.get("/api/user", headers=headers)
            self.assertEqual(result.status_code, 200)

    def test_user_post(self):
        headers = self.login("root")
        new_user = {
            "username": "unittest",
            "firstname": "unit",
            "lastname": "test",
            "roles": ["admin", "root"],
            "password": "super-secret"
        }
        result = self.app.post("/api/user", headers=headers,
            json=new_user)
        self.assertEqual(result.status_code, 201)

        result = self.app.post("/api/user", headers=headers,
            json=new_user)
        self.assertEqual(result.status_code, 400)

    def test_user_delete(self):
        headers = self.login("root")
        result = self.app.delete("/api/user/1", headers=headers)
        self.assertEqual(result.status_code, 200)

    def test_user_delete_unknown(self):
        headers = self.login("root")
        result = self.app.delete("/api/user/99999", headers=headers)
        self.assertEqual(result.status_code, 404)

    def test_user_delete_forbidden(self):
        headers = self.login("user")
        result = self.app.delete("/api/user/4", headers=headers)
        self.assertEqual(result.status_code, 403)

    def test_user_patch(self):
        headers = self.login("root")
        result = self.app.patch("/api/user/1", headers=headers, json={
            "username": "root2",
            "firstname": "henk",
            "lastname": "biertje",
            "password": "wachtwoord",
            "roles": ["root"]
        })
        self.assertEqual(result.status_code, 200)

    def test_user_patch_unknown(self):
        headers = self.login("root")
        result = self.app.patch("/api/user/9999", headers=headers, json={
            "username": "root2"
        })
        self.assertEqual(result.status_code, 404)

    def test_user_patch_forbidden(self):
        headers = self.login("user")
        result = self.app.patch("/api/user/4", headers=headers, json={
            "username": "root2"
        })
        self.assertEqual(result.status_code, 403)

    def test_root_username_forbidden(self):
        headers = self.login("root")
        results = self.app.post("/api/user", headers=headers, json={
            "username": "root",
            "firstname": "madman",
            "lastname": "idiot",
            "roles": "admin",
            "password": "something-really-secure"
        })
        self.assertEqual(results.status_code, 400)

    def test_root_role_forbidden(self):
        headers = self.login("root")
        new_user = {
            "username": "some",
            "firstname": "guy",
            "lastname": "there",
            "roles":  "root",
            "password": "super-secret"
        }
        result = self.app.post("/api/user", headers=headers,
            json=new_user)
        self.assertEqual(result.status_code, 400)



import logging

from vantage6.server import db
from vantage6 import server
from flask_jwt_extended import verify_jwt_in_request, get_jwt_identity
import jwt
from flask_socketio import join_room, send, leave_room, emit, Namespace
from flask import g, session, request

class DefaultSocketNamespace(Namespace):
    """Handlers for SocketIO events are different than handlers for routes and that
    introduces a lot of confusion around what can and cannot be done in a SocketIO handler.
    The main difference is that all the SocketIO events generated for a client occur in
    the context of a single long running request.
    """

    log = logging.getLogger(__name__)

    def on_connect(self):
        """New incomming connections are authenticated using their
        JWT authorization token which is obtained from the REST api.
        A session is created for each connected clients which lives
        as long as the connection is active. There has not been made
        any difference between connecting and re-connecting.
        """

        self.log.info(f'Client connected: "{request.sid}"')

        # try to catch jwt authorization token.
        try:
            verify_jwt_in_request()

        except jwt.exceptions.ExpiredSignatureError as e:
            self.log.error("JWT has expired")
            emit("expired_token", "", room=request.sid)

        except Exception as e:
            self.log.error("Could not connect client! No or Invalid JWT token?")
            self.log.exception(e)
            session.name = "no-sure-yet"
            self.__join_room_and_notify(request.sid)

            # FIXME: expired probably doesn't cover it ...
            emit("expired_token", "", room=request.sid)
            return

        # get identity from token.
        user_or_node_id = get_jwt_identity()
        session.auth = auth = db.Authenticatable.get(user_or_node_id)
        session.auth.status = 'online'
        session.auth.save()

        # It appears to be necessary to use the root socketio instance
        # otherwise events cannot be sent outside the current namespace.
        # In this case, only events to '/tasks' can be emitted otherwise.
        server.socketio.emit('node-status-changed', namespace='/admin')

        # define socket-session variables.
        session.type = auth.type
        session.name = auth.username if session.type == 'user' else auth.name
        self.log.info(f'Client identified as <{session.type}>: <{session.name}>')

        # join appropiate rooms, nodes join a specific collaboration room.
        # users do not belong to specific collaborations.
        session.rooms = ['all_connections', 'all_'+session.type+'s']

        if session.type == 'node':
            session.rooms.append('collaboration_' + str(auth.collaboration_id))
            session.rooms.append('node_' + str(auth.id))
        elif session.type == 'user':
             session.rooms.append('user_'+str(auth.id))

        for room in session.rooms:
            self.__join_room_and_notify(room)


    def on_disconnect(self):
        # FIXME: this raises errors:
        #   AttributeError: '_ManagedSession' object has no attribute '...'
        #
        # for room in session.rooms:
        #     self.__leave_room_and_notify(room)
        #
        # session.auth.status = 'offline'
        # session.auth.save()

        # It appears to be necessary to use the root socketio instance
        # otherwise events cannot be sent outside the current namespace.
        # In this case, only events to '/tasks' can be emitted otherwise.
        self.log.warning('emitting to /admin')
        server.socketio.emit('node-status-changed', namespace='/admin')

        self.log.info(f'{session.name} disconnected')

    def on_message(self, message):
        self.log.info('received message: ' + message)

    def on_error(self, e):
        self.log.error(e)

    def on_join_room(self, room):
        self.__join_room_and_notify(room)

    def on_container_failed(self, node_id, status_code, result_id, collaboration_id):
        run_id = db.Result.get(result_id).task.run_id

        self.log.critical(
            f"A container in for run_id={run_id} and result_id={result_id}"
            f" within collaboration_id={collaboration_id} on node_id={node_id}"
            f" exited with status_code={status_code}."
        )
        # emit('message', "somewhere in the universe a container has crashed", room='all_connections')
        # print("collaboration_"+str(collaboration_id))

        room = "collaboration_"+str(collaboration_id)
        emit("container_failed", run_id, room=room)

    def on_ping(self, node_id):
        # self.log.debug(f"ping from id={node_id}")
        room = f"node_{node_id}"
        emit("pang","success!", room=room)

    def __join_room_and_notify(self, room):
        join_room(room)
        msg = f'<{session.name}> joined room <{room}>'
        self.log.info(msg)
        emit('message', msg, room=room)

    def __leave_room_and_notify(self, room):
        leave_room(room)
        msg = f'{session.name} left room {room}'
        self.log.info(msg)
        emit('message', msg, room=room)





from pathlib import Path
from vantage6.common.globals import APPNAME, STRING_ENCODING

#
#   INSTALLATION SETTINGS
#
PACAKAGE_FOLDER = Path(__file__).parent.parent.parent

DATA_FOLDER = PACAKAGE_FOLDER / APPNAME / "server" / "_data"

# with open(Path(PACAKAGE_FOLDER) / APPNAME / "server" / "VERSION") as f:
#     VERSION = f.read()


from vantage6.common import Singleton, logger_name, bytes_to_base64s, base64s_to_bytes


def log_full_request(request, log=None):
    if log is None:
        stack = inspect.stack()
        calling = stack[1]
        filename = os.path.split(calling.filename)[-1]
        module_name = os.path.splitext(filename)[0]

        log = logging.getLogger(module_name)

    log.info(f'{request.method}: {request.url}')
    log.info(f'  request.args: {request.args}')
    log.info(f'  request.data: {request.get_data()}')

    if request.is_json and len(request.data):
        log.info(f'  request.json: {request.json}')

    log.info(f'  headers:')
    for header in str(request.headers).splitlines():
        log.info(f'    ' + header)



# -*- coding: utf-8 -*-
"""
Resources below '/<api_base>/result'
"""
import logging

from flask import g, request, url_for
from flask_restful import Resource, abort
from requests import codes as rqc # todo remove and use HTTPStatus
from http import HTTPStatus
from . import parse_datetime
from . import with_user_or_node, with_node, only_for
from ._schema import *
from flasgger import swag_from
from http import HTTPStatus
from pathlib import Path

from vantage6.server import socketio, api
from vantage6.server.model import ( Result as db_Result, Node,
    Task, Collaboration, Organization )
from vantage6.server.model.base import Database

module_name = __name__.split('.')[-1]
log = logging.getLogger(module_name)


def setup(api, api_base):

    path = "/".join([api_base, module_name])
    log.info('Setting up "{}" and subdirectories'.format(path))

    api.add_resource(
        Result,
        path,
        endpoint='result_without_id',
        methods=('GET',)
    )
    api.add_resource(
        Result,
        path + '/<int:id>',
        endpoint='result_with_id',
        methods=('GET', 'PATCH')
    )


# Schemas
result_schema = ResultSchema()
result_inc_schema = ResultTaskIncludedSchema()


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class Result(Resource):
    """Resource for /api/result"""

    @only_for(['node', 'user', 'container'])
    @swag_from(str(Path(r"swagger/get_result_with_id.yaml")),endpoint="result_with_id")
    @swag_from(str(Path(r"swagger/get_result_without_id.yaml")), endpoint="result_without_id")
    def get(self, id=None):
        if id:
            t = db_Result.get(id)
        else:

            session = Database().Session
            q = session.query(db_Result)

            if request.args.get('state') == 'open':
                q = q.filter(db_Result.finished_at == None)

            # q = q.join(db_Result)
            if request.args.get('task_id'):
             q = q.filter_by(task_id=request.args.get('task_id'))

            q = q.join(Organization).join(Node).join(Task, db_Result.task).join(Collaboration)
            if request.args.get('node_id'):
                q = q.filter(db.Node.id==request.args.get('node_id'))\
                    .filter(db.Collaboration.id==db.Node.collaboration_id)

            t = q.all()

        if request.args.get('include') == 'task':
            s = result_inc_schema
        else:
            s = result_schema

        return s.dump(t, many=not bool(id)).data, HTTPStatus.OK

    @with_node
    @swag_from(str(Path(r"swagger/patch_result_with_id.yaml")), endpoint="result_with_id")
    def patch(self, id):
        """Update a Result."""
        data = request.get_json()
        result = db_Result.get(id)

        if result.organization_id != g.node.organization_id:
            log.warn(
                f"{g.node.name} tries to update a result that does not belong "
                f"to him. ({result.organization_id}/{g.node.organization_id})"
            )
            return {"msg": "This is not your result to PATCH!"}, HTTPStatus.UNAUTHORIZED

        if result.finished_at is not None:
            return {"msg": "Cannot update an already finished result!"}, HTTPStatus.BAD_REQUEST


        # notify collaboration nodes/users that the task has an update
        socketio.emit(
            "status_update",
            {'result_id': id},
            room='collaboration_'+str(result.task.collaboration.id),
            namespace='/tasks',
        )

        url = url_for('result_with_id', id=id)
        # log.debug(f'result [{url}] was updated.')

        result.started_at = parse_datetime(data.get("started_at"), result.started_at)
        result.finished_at = parse_datetime(data.get("finished_at"))
        result.result = data.get("result")
        result.log = data.get("log")

        result.save()

        return result



import bcrypt

from sqlalchemy import Column, String, Integer, ForeignKey, exists
from sqlalchemy.orm import relationship

from .base import Database
from .authenticable import Authenticatable


class User(Authenticatable):
    """User (person) that can access the system.

    Users always belong to an organization and can have certain
    rights within an organization.
    """
    _hidden_attributes = ['password']

    # overwrite id with linked id to the authenticatable
    id = Column(Integer, ForeignKey('authenticatable.id'), primary_key=True)
    __mapper_args__ = {
        'polymorphic_identity':'user',
    }

    # fields
    username = Column(String, unique=True)
    password = Column(String)
    firstname = Column(String)
    lastname = Column(String)
    roles = Column(String)
    organization_id = Column(Integer, ForeignKey("organization.id"))

    # relationships
    organization = relationship("Organization", back_populates="users")

    def __repr__(self):
        organization = self.organization.name if self.organization else "None"
        return (
            f"<User "
            f"id={self.id}, username='{self.username}', roles='{self.roles}', "
            f"organization='{organization}'"
            f">"
        )

    def set_password(self, pw):
        pwhash = bcrypt.hashpw(pw.encode('utf8'), bcrypt.gensalt())
        self.password = pwhash.decode('utf8')

    def check_password(self, pw):
        if self.password is not None:
            expected_hash = self.password.encode('utf8')
            return bcrypt.checkpw(pw.encode('utf8'), expected_hash)
        return False

    @classmethod
    def getByUsername(cls, username):
        session = Database().Session
        return session.query(cls).filter_by(username=username).one()

    @classmethod
    def get_user_list(cls, filters=None):
        session = Database().Session
        return session.query(cls).all()

    @classmethod
    def username_exists(cls, username):
        session = Database().Session
        return session.query(exists().where(cls.username == username)).scalar()



# -*- coding: utf-8 -*-
"""
Resources below '/<api_base>/user'
"""
from __future__ import print_function, unicode_literals

import logging

from flask_restful import Resource
from flask import g
from flask_restful import reqparse
from http import HTTPStatus
from flasgger import swag_from
from pathlib import Path

from vantage6.server import db
from vantage6.server.resource import with_user, only_for
from vantage6.server.resource._schema import UserSchema

module_name = __name__.split('.')[-1]
log = logging.getLogger(module_name)


def setup(api, API_BASE):
    module_name = __name__.split('.')[-1]
    path = "/".join([API_BASE, module_name])
    log.info('Setting up "{}" and subdirectories'.format(path))

    api.add_resource(
        User,
        path,
        endpoint='user_without_id',
        methods=('GET', 'POST')
    )
    api.add_resource(
        User,
        path + '/<int:id>',
        endpoint='user_with_id',
        methods=('GET', 'PATCH', 'DELETE')
    )


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class User(Resource):

    user_schema = UserSchema()

    @only_for(['user'])
    @swag_from(str(Path(r"swagger/get_user_with_id.yaml")), endpoint='user_with_id')
    @swag_from(str(Path(r"swagger/get_user_without_id.yaml")), endpoint='user_without_id')
    def get(self, id=None):
        """Return user details."""
        retval = db.User.get(id)
        is_root = g.user.username == "root"

        if id is None:
            # Only root can retrieve all users at once
            if is_root:
                return self.user_schema.dump(retval, many=True)

            # Everyone else can only list the users from their own organization
            else:
                org_id = g.user.organization_id
                filtered = [u for u in retval if u.organization_id == org_id]
                return self.user_schema.dump(filtered, many=True)

        if retval:
            # You either have to be root or someone from the same organization
            if is_root or (retval.organization_id == g.user.organization_id):
                return self.user_schema.dump(retval, many=False)

        return {"msg": f"user id {id} is not found"}, HTTPStatus.NOT_FOUND

    @with_user
    @swag_from(str(Path(r"swagger/post_user_without_id.yaml")), endpoint='user_without_id')
    def post(self):
        """Create a new User."""

        parser = reqparse.RequestParser()
        parser.add_argument("username", type=str, required=True, help="This field is required")
        parser.add_argument("firstname", type=str, required=True, help="This field is required")
        parser.add_argument("lastname", type=str, required=True, help="This field is required")
        parser.add_argument("roles", type=str, required=True, help="This field is required")
        parser.add_argument("password", type=str, required=True, help="This field is required")
        data = parser.parse_args()

        if db.User.username_exists(data["username"]):
            return {"msg": "username already exists"}, HTTPStatus.BAD_REQUEST

        if data["username"] == 'root':
            msg = {"msg": "You're funny! You can't create root!?"}
            return msg, HTTPStatus.BAD_REQUEST

        roles = data['roles'].split(',')
        if 'root' in roles:
            msg = {"msg": "You're funny! You can't assign the role 'root'!?"}
            return msg, HTTPStatus.BAD_REQUEST

        # Ok, looks like we got most of the security hazards out of the way
        user = db.User(
            username=data["username"],
            firstname=data["firstname"],
            lastname=data["lastname"],
            roles=data["roles"],
            organization_id=g.user.organization_id
        )

        user.set_password(data["password"])
        user.save()

        return self.user_schema.dump(user), HTTPStatus.CREATED
    @with_user
    @swag_from(str(Path(r"swagger/patch_user_with_id.yaml")), endpoint='user_with_id')
    def patch(self, id):

        user = db.User.get(id)

        if not user:
            return {"msg": "user id={} not found".format(id)}, HTTPStatus.NOT_FOUND
        if user.organization_id != g.user.organization_id and g.user.roles != "admin":
            return {"msg": "you do not have permission to modify user id={}".format(id)}, HTTPStatus.FORBIDDEN

        parser = reqparse.RequestParser()
        parser.add_argument("username", type=str, required=False, help="This field is required")
        parser.add_argument("firstname", type=str, required=False, help="This field is required")
        parser.add_argument("lastname", type=str, required=False, help="This field is required")
        parser.add_argument("roles", type=str, required=False, help="This field is required")
        parser.add_argument("password", type=str, required=False, help="This field is required")
        data = parser.parse_args()

        if data["username"]:
            user.username = data["username"]
        if data["firstname"]:
            user.firstname = data["firstname"]
        if data["lastname"]:
            user.lastname = data["lastname"]
        if data["password"]:
            user.set_password(data["password"])
        if data["roles"]:
            user.roles = data["roles"]

        user.save()
        return user, HTTPStatus.OK

    @with_user
    @swag_from(str(Path(r"swagger/delete_user_with_id.yaml")), endpoint='user_with_id')
    def delete(self, id):

        user = db.User.get(id)
        if not user:
            return {"msg": "user id={} not found".format(id)}, HTTPStatus.NOT_FOUND
        if user.organization_id != g.user.organization_id and g.user.roles != "admin":
            log.warning(
                "user {} has tried to delete user {} but does not have the required permissions".format(
                    g.user.id,
                    user.id
                )
            )
            return {"msg": "you do not have permission to modify user id={}".format(id)}, HTTPStatus.FORBIDDEN

        user.delete()
        log.info("user id={} is removed from the database".format(id))
        return {"msg": "user id={} is removed from the database".format(id)}, HTTPStatus.OK



# swagger template
swagger_template = {
    "components":{
        "securitySchemes":{
            "bearerAuth":{
                "type": "http",
                "scheme": "bearer",
                "bearerFormat": " JWT"
            }
        },
        "schemas":{
            "Task":{
                "properties":{
                    "image": {"type": "string"},
                    "description": {"type": "string"},
                    "input": {"type": "string"},
                    "name": {"type": "string"},
                    "collaboration_id": {"type": "integer"},
                    "organization_ids": {
                        "type": "array",
                        "items": {"type": "integer"}
                    }
                },
                "example":{
                    "name": "human-readable-name",
                    "image": "hello-world",
                    "collaboration_id": 1
                },
                "required": ["image", "collaboration_id"]
            },
            "Collaboration": {
                "properties": {
                    "collaboration_id": {"type": "integer"}
                },
                "example":{
                    "collaboration_id": 1
                }
            },
            "Node": {
                "example": {
                    "api_key": "unique-string"
                },
                "properties": {
                    "api_key": {
                        "type": "string"
                    }
                }
            },
            "User": {
                "example": {
                    "password": "secret!", 
                    "username": "yourname"
                }, 
                "properties": {
                    "password": {"type": "string"}, 
                    "username": {"type": "string"}
                }
            },
            "ContainerToken": {
                "properties": {
                    "task_id": {"type": "string"},
                    "image": {"type": "string"}
                },
                "example": {
                    "task_id": 1,
                    "image": "hello-world"
                }
            }
        }
    },
    "security": [
        {"bearerAuth": []}
    ]
}


# -*- coding: utf-8 -*-
"""
Resources below '/<api_base>/version'
"""
from __future__ import print_function, unicode_literals

import logging

from flask_restful import Resource, abort
from flasgger import swag_from
from pathlib import Path

# from vantage6.server.globals import VERSION
import vantage6.server


module_name = __name__.split('.')[-1]
log = logging.getLogger(module_name)

from .. import db

def setup(api, API_BASE):

    path = "/".join([API_BASE, module_name])
    log.info('Setting up "{}" and subdirectories'.format(path))

    api.add_resource(
        Version,
        path,
        endpoint='version',
        methods=('GET',)
    )


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class Version(Resource):

    @swag_from(str(Path(r"swagger/version.yaml")), endpoint='version')
    def get(self):
        """Return the version of this server."""

        return {"version": vantage6.server.__version__}


from sqlalchemy import Column, String, DateTime

from .base import Base


class Authenticatable(Base):
    """Parent table of all entities that can authenticate. 
    
    Entities that can authenticate are nodes and users. Containers
    can also authenticate but these are authenticated inderect 
    through the nodes.
    """

    # fields
    type = Column(String(50))
    __mapper_args__ = {
        'polymorphic_identity': 'authenticatable',
        'polymorphic_on': type,
    }
    ip = Column(String)
    last_seen = Column(DateTime)
    status = Column(String)
